"""DataTracer functional and REST api.

This module implements a set of simple python functions
to expose the different types of solvers from DataTracer
in a simplified functional interface.

This functional interface is also exposed through a REST
API based on [hug](https://hug.rest).
"""

import io
import uuid
from copy import deepcopy

import hug
import pandas as pd
from hug import types

from datatracer import DataTracer

PRIMARY_KEY_SOLVER = 'datatracer.primary_key.basic'
FOREIGN_KEY_SOLVER = 'datatracer.foreign_key.standard'
COLUMN_MAP_SOLVER = 'datatracer.column_map.basic'
COLUMN_MAP_THRESHOLD = 0.1


def _load_table_data(table):
    data = table.pop('data', None)
    if data is not None:
        if isinstance(data, pd.DataFrame):
            return data

        with io.StringIO(data) as sio:
            return pd.read_csv(sio)

    path = table.get('path')
    if path is None:
        raise ValueError('Either `data` or `path` must be provided')

    return pd.read_csv(path)


def _get_table_id(details, table):
    selected = []
    for table_id, table_details in details.items():
        for key, value in table.items():
            if key not in table_details or value != table_details[key]:
                continue

            selected.append(table_id)

    if len(selected) > 1:
        raise ValueError('More than one table matches the given table identifier')
    elif not selected:
        raise ValueError('No table matches the given table identifier')

    return selected[0]


def load_tables(tables):
    data = {}
    details = {}
    for table in tables:
        table_id = uuid.uuid4()
        table_details = deepcopy(table)
        data[table_id] = _load_table_data(table_details)
        details[table_id] = table_details

    return data, details


@hug.post('/primary_key_detection')
def primary_key_detection(tables: types.multiple):
    """Solve a primary key detection problem."""
    tables, details = load_tables(tables)

    solver = DataTracer.load(PRIMARY_KEY_SOLVER)
    primary_keys = solver.solve(tables)

    response = []
    for table_id, primary_key in primary_keys.items():
        table_details = details[table_id]
        table_details['primary_key'] = primary_key
        response.append(table_details)

    return {
        'primary_keys': response
    }


@hug.post('/foreign_key_detection')
def foreign_key_detection(tables: types.multiple):
    """Solve a foreign key detection problem."""
    tables, details = load_tables(tables)

    solver = DataTracer.load(FOREIGN_KEY_SOLVER)
    foreign_keys = solver.solve(tables)

    for foreign_key in foreign_keys:
        foreign_key['table'] = details[foreign_key['table']]
        foreign_key['ref_table'] = details[foreign_key['ref_table']]

    return {
        'foreign_keys': foreign_keys
    }


@hug.post('/column_mapping')
def column_mapping(tables: types.multiple, target_table: types.json, target_field: types.text):
    """Solve a column mapping problem."""
    tables, details = load_tables(tables)
    target_table_id = _get_table_id(details, target_table)

    solver = DataTracer.load(COLUMN_MAP_SOLVER)
    column_map = solver.solve(tables, target_table=target_table_id, target_field=target_field)

    column_mappings = []
    for (table_id, field), score in column_map.items():
        if score >= COLUMN_MAP_THRESHOLD:
            column_mappings.append({
                'table': details[table_id],
                'field': field,
            })

    return {
        'target_table': target_table,
        'target_field': target_field,
        'column_mappings': column_mappings
    }



import logging

from sklearn.ensemble import RandomForestRegressor

from datatracer.column_map.base import ColumnMapSolver
from datatracer.column_map.transformer import Transformer

LOGGER = logging.getLogger(__name__)


class BasicColumnMapSolver(ColumnMapSolver):
    """Basic Solver for the data lineage problem of column dependency."""

    def __init__(self, *args, **kwargs):
        self._model_args = args
        self._model_kwargs = kwargs

    def _get_importances(self, X, y):
        model = RandomForestRegressor(*self._model_args, **self._model_kwargs)
        model.fit(X, y)

        return model.feature_importances_

    def solve(self, tables, foreign_keys, target_table, target_field):
        """Find the fields which contributed to the target_field the most.

        The output is a dictionary containing the fields that contributed the
        most to the given target field as keys, specified as a tuple containing
        both table name and field name, and the score obtained as values.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.
            foreign_keys (list):
                List of foreign key specifications.
            target_table (str):
                Name of the table that contains the target field.
            target_field (str):
                Name of the target field.

        Returns:
            dict:
                Dictionary of field specification tuples and scores.
        """
        transformer = Transformer(tables, foreign_keys)

        X, y = transformer.forward(target_table, target_field)

        importances = self._get_importances(X, y)
        return transformer.backward(importances)



"""Data module.

This module provides data to load datasets as tuples of
metadata and tables, as well as a helper function to
generate a folder with demo datasets inside it.
"""
import os
import shutil

import pandas as pd
from metad import MetaData


def load_dataset(dataset_path):
    """Load a dataset as a MetaData and a dict of tables.

    Args:
        dataset_path (str):
            Path to the root of the dataset.

    Returns:
        tuple:
            Tuple containing:
                * A ``MetaData`` instance.
                * A ``dict`` containing the tables loaded as ``pandas.DataFrames``.
    """
    metadata = MetaData.from_json(os.path.join(dataset_path, 'metadata.json'))
    tables = {}
    for table_name in metadata.get_table_names():
        tables[table_name] = pd.read_csv(os.path.join(dataset_path, table_name + '.csv'))

    return metadata, tables


def load_datasets(datasets_path):
    """Load a collection of datasets.

    The datasets are loaded as tuples containing a MetaData instance
    and a dict with the tables of the dataset loaded as DataFrames.

    Args:
        datasets_path (str):
            Path to the folder where the dataset can be found.

    Returns:
        dict:
            Dict of (metadata, tables) tubles, one for each dataset.
    """
    return {
        dataset: load_dataset(os.path.join(datasets_path, dataset))
        for dataset in os.listdir(datasets_path)
    }


def get_demo_data(path='datatracer_demo', force=False):
    """Get a folder with demo data inside it.

    By default, the demo folder will be called ``datatracer_demo`` and will
    be placed inside the current working directory. Optionally, a different
    path can be passed.

    If the destination path exists and force is True it will be removed.
    Otherwise, an error will be raised.

    Args:
        path (str):
            Destination path where the folder will be created, including the
            folder name. Defaults to ``datatracer_demo``.
        force (bool):
            Whether to delete the destination folder if it already exists.
            Defaults to ``False``.

    Raises:
        FileExistsError:
            If the destination folder already exists and force is False.
    """
    if os.path.exists(path):
        if force:
            shutil.rmtree(path)
        else:
            msg = 'Path "{}" already exists. Please remove it or use `force=True`'.format(path)
            raise FileExistsError(msg)
    else:
        os.makedirs(os.path.dirname(os.path.abspath(path)), exist_ok=True)

    if os.path.isabs(path):
        print('Generating a demo folder at `{}`'.format(path))
    else:
        print('Generating a demo folder at `./{}`'.format(path))
        path = os.path.abspath(path)

    demo_path = os.path.join(os.path.dirname(__file__), 'datasets')
    shutil.copytree(demo_path, path)



# -*- coding: utf-8 -*-

"""Top-level package for DataTracer."""

__author__ = 'MIT Data To AI Lab'
__email__ = 'dailabmit@gmail.com'
__version__ = '0.0.4'

import os

from mlblocks import discovery

from datatracer.core import PRETRAINED_DIR, DataTracer
from datatracer.data import get_demo_data, load_dataset, load_datasets

_BASE_PATH = os.path.abspath(os.path.dirname(__file__))
_JSONS_PATH = os.path.join(_BASE_PATH, 'jsons')
MLBLOCKS_PRIMITIVES = os.path.join(_JSONS_PATH, 'primitives')
MLBLOCKS_PIPELINES = os.path.join(_JSONS_PATH, 'pipelines')


__all__ = (
    'DataTracer',
    'get_demo_data',
    'get_pipelines',
    'get_primitives',
    'load_dataset',
    'load_datasets',
)


def get_pipelines():
    """Get a list of the available datatracer pipelines.

    Returns:
        list:
            List of the names of the available datatracer pipelines.
    """
    return discovery.find_pipelines('datatracer')


def get_primitives():
    """Get a list of the available datatracer primitives.

    Returns:
        list:
            List of the names of the available datatracer primitives.
    """
    return discovery.find_primitives('datatracer')


def get_solvers():
    """Get a list of the available datatracer solvers.

    Returns:
        list:
            List of the names of the available datatracer solvers.
    """
    return [
        pretrained_solver[:-3]
        for pretrained_solver in os.listdir(PRETRAINED_DIR)
        if pretrained_solver.endswith('.dt')
    ]



from collections import Counter
from itertools import permutations

from sklearn.ensemble import RandomForestClassifier
from tqdm import tqdm

from datatracer.foreign_key.base import ForeignKeySolver


class StandardForeignKeySolver(ForeignKeySolver):

    def __init__(self, threshold=0.9, add_details=False, *args, **kwargs):
        self._threshold = threshold
        self._add_details = add_details
        self._model_args = args
        self._model_kwargs = kwargs

    def _diff(self, a, b):
        diff = 0
        a, b = Counter(a), Counter(b)
        for k in set(a.keys()).union(set(b.keys())):
            diff += abs(a[k] - b[k])

        return diff

    def _feature_vector(self, parent_col, child_col):
        parent_set, child_set = set(parent_col), set(child_col)
        len_intersect = len(parent_set.intersection(child_set))
        return [
            len_intersect / (len(child_set) + 1e-5),
            len_intersect / (len(parent_set) + 1e-5),
            len_intersect / (max(len(child_set), len(parent_set)) + 1e-5),
            1.0 if parent_col.name == child_col.name else 0.0,
            self._diff(parent_col.name, child_col.name),
            1.0 if child_set.issubset(parent_set) else 0.0,
            1.0 if parent_set.issubset(child_set) else 0.0,
            len(parent_set) / (len(parent_col) + 1e-5),
            1.0 if "key" in parent_col.name else 0.0,
            1.0 if "id" in parent_col.name else 0.0,
            1.0 if "_key" in parent_col.name else 0.0,
            1.0 if "_id" in parent_col.name else 0.0,
            1.0 if parent_col.dtype == "int64" else 0.0,
            1.0 if parent_col.dtype == "object" else 0.0,
            len(child_set) / (len(child_col) + 1e-5),
            1.0 if "key" in child_col.name else 0.0,
            1.0 if "id" in child_col.name else 0.0,
            1.0 if "_key" in child_col.name else 0.0,
            1.0 if "_id" in child_col.name else 0.0,
            1.0 if child_col.dtype == "int64" else 0.0,
            1.0 if child_col.dtype == "object" else 0.0,
        ]

    def fit(self, list_of_databases):
        """Fit this solver.

        Args:
            list_of_databases (list):
                List of tuples containing ``MetaData`` instnces and table dictinaries,
                which contain table names as input and ``pandas.DataFrames`` as values.
        """
        X, y = [], []
        for metadata, tables in tqdm(list_of_databases, "extracting features"):
            fks = metadata.get_foreign_keys()
            fks = set([
                (fk["table"], fk["field"], fk["ref_table"], fk["ref_field"])
                for fk in fks
            ])

            for t1, t2 in permutations(tables.keys(), r=2):
                for c1 in tables[t1].columns:
                    for c2 in tables[t2].columns:
                        if tables[t1][c1].dtype.kind != tables[t2][c2].dtype.kind:
                            continue

                        if len(tables[t1][c1]) == 0 or len(tables[t2][c2]) == 0:
                            print(tables.keys())
                            raise RuntimeError("Found empty table!")

                        X.append(self._feature_vector(tables[t2][c2], tables[t1][c1]))
                        y.append(1.0 if (t1, c1, t2, c2) in fks else 0.0)

        self.model = RandomForestClassifier(*self._model_args, **self._model_kwargs)
        self.model.fit(X, y)

    def solve(self, tables, primary_keys=None):
        """Solve the foreign key detection problem.

        The output is a list of foreign key specifications, in order from the most likely
        to the least likely.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.
            primary_keys (dict):
                (Ignored). This particular implementation does not use this argument.

        Returns:
            dict:
                List of foreign key specifications, sorted by likelyhood.
        """
        X, foreign_keys = [], []
        for t1, t2 in permutations(tables.keys(), r=2):
            for c1 in tables[t1].columns:
                for c2 in tables[t2].columns:
                    if tables[t1][c1].dtype.kind != tables[t2][c2].dtype.kind:
                        continue

                    X.append(self._feature_vector(tables[t2][c2], tables[t1][c1]))
                    foreign_keys.append((t1, c1, t2, c2))

        if not foreign_keys:
            return []

        scores = self.model.predict(X)

        best_foreign_keys = []
        for score, (t1, c1, t2, c2) in sorted(zip(scores, foreign_keys), reverse=True):
            if self._threshold is None or score >= self._threshold:
                foreign_key = {
                    "table": t1,
                    "field": c1,
                    "ref_table": t2,
                    "ref_field": c2,
                }
                if self._add_details:
                    foreign_key.update({
                        "score": score,
                        "features": self._feature_vector(tables[t2][c2], tables[t1][c1])
                    })

                best_foreign_keys.append(foreign_key)

        return best_foreign_keys



"""Primary Key Solving base class."""


class PrimaryKeySolver():

    def fit(self, list_of_databases):
        """Fit this solver.

        Args:
            list_of_databases (list):
                List of tuples containing ``MetaData`` instnces and table dictinaries,
                which contain table names as input and ``pandas.DataFrames`` as values.
        """
        pass

    def solve(self, tables):
        """Solve the primary key detection problem.

        The output is a dictionary contiaining table names as keys, and the
        name of the field that is most likely to be the primary key as values.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.

        Returns:
            dict:
                Dict containing table names as keys and field names as values.
        """
        raise NotImplementedError()



"""Basic Foreign Key Solver module."""

from itertools import permutations

from tqdm import tqdm

from datatracer.foreign_key.base import ForeignKeySolver


class BasicForeignKeySolver(ForeignKeySolver):

    def __init__(self, threshold=0.9, add_details=False):
        self._threshold = threshold
        self._add_details = add_details

    def _score(self, col_a, col_b):
        set_a, set_b = set(col_a), set(col_b)
        if set_b.issubset(set_a):  # child must be subset of parent
            num = len(set_a.intersection(set_b))
            denom = max(len(set_a), len(set_b))
            return num / (denom + 1e-5)

        return 0.0

    def solve(self, tables, primary_keys=None):
        """Solve the foreign key detection problem.

        The output is a list of foreign key specifications, in order from the most likely
        to the least likely.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.
            primary_keys (dict):
                (Ignored). This particular implementation does not use this argument.

        Returns:
            dict:
                List of foreign key specifications, sorted by likelyhood.
        """
        foreign_keys = []
        for t1, t2 in tqdm(list(permutations(tables.keys(), r=2))):
            for c1 in tables[t1].columns:
                for c2 in tables[t2].columns:
                    if tables[t1][c1].dtype.kind != tables[t2][c2].dtype.kind:
                        continue

                    score = self._score(tables[t1][c1], tables[t2][c2])
                    foreign_keys.append((score, t1, c1, t2, c2))

        best_foreign_keys = []
        for score, t1, c1, t2, c2 in sorted(foreign_keys, reverse=True):
            if self._threshold is None or score >= self._threshold:
                foreign_key = {
                    "table": t1,
                    "field": c1,
                    "ref_table": t2,
                    "ref_field": c2,
                }
                if self._add_details:
                    foreign_key['score'] = score

                best_foreign_keys.append(foreign_key)

        return best_foreign_keys



"""Basic Primary Key Solver module."""

import numpy as np
from sklearn.ensemble import RandomForestClassifier

from datatracer.primary_key.base import PrimaryKeySolver


class BasicPrimaryKeySolver(PrimaryKeySolver):

    def __init__(self, *args, **kwargs):
        self._model_args = args
        self._model_kwargs = kwargs

    def _feature_vector(self, table, column_name):
        column = table[column_name]
        return [
            list(table.columns).index(column_name),
            list(table.columns).index(column_name) / len(table.columns),
            1.0 if len(set(column)) == len(column) else 0.0,
            len(set(column)) / len(column),
            1.0 if "key" in column.name else 0.0,
            1.0 if "id" in column.name else 0.0,
            1.0 if "_key" in column.name else 0.0,
            1.0 if "_id" in column.name else 0.0,
            1.0 if column.dtype == "int64" else 0.0,
            1.0 if column.dtype == "object" else 0.0,
        ]

    def fit(self, list_of_databases):
        """Fit this solver.

        Args:
            list_of_databases (list):
                List of tuples containing ``MetaData`` instnces and table dictinaries,
                which contain table names as input and ``pandas.DataFrames`` as values.
        """
        X, y = [], []
        for metadata, tables in list_of_databases:
            for table in metadata.get_tables():
                if "primary_key" not in table:
                    continue
                if not isinstance(table["primary_key"], str):
                    continue

                primary_key = table["primary_key"]
                for column in tables[table["name"]].columns:
                    X.append(self._feature_vector(tables[table["name"]], column))
                    y.append(1.0 if primary_key == column else 0.0)

        X, y = np.array(X), np.array(y)

        self.model = RandomForestClassifier(*self._model_args, **self._model_kwargs)
        self.model.fit(X, y)

    def _find_primary_key(self, table):
        best_column, best_score = None, float("-inf")
        for column in table.columns:
            score = self.model.predict([self._feature_vector(table, column)])
            if score > best_score:
                best_column = column
                best_score = score

        return best_column

    def solve(self, tables):
        """Solve the problem.

        The output is a dictionary contiaining table names as keys, and the
        name of the field that is most likely to be the primary key as values.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.

        Returns:
            dict:
                Dict containing table names as keys and field names as values.
        """
        primary_keys = {}
        for table in tables:
            primary_keys[table] = self._find_primary_key(tables[table])

        return primary_keys



import argparse
import json
import logging
import sys

import hug
import yaml

from datatracer import api

API_CONFIG_KEYS = (
    'host',
    'port',
    'primary_key_solver',
    'foreign_key_solver',
    'column_map_solver',
    'column_map_threshold',
)


def print_box(message, strong=False):
    char = '#' if strong else '*'
    line = char * max(len(line) for line in message.split('\n'))
    print('\n'.join((line, message, line)))


def _api_serve(args):
    if args.config:
        with open(args.config) as config_file:
            if args.config.endswith('yaml') or args.config.endswith('yml'):
                config = yaml.safe_load(config_file)
            else:
                config = json.load(config_file)

            for key in config:
                if key not in API_CONFIG_KEYS:
                    raise ValueError('Unknown option `{}` in {}'.format(key, args.config))

        api.PRIMARY_KEY_SOLVER = config.get('primary_key_solver', api.PRIMARY_KEY_SOLVER)
        api.FOREIGN_KEY_SOLVER = config.get('foreign_key_solver', api.FOREIGN_KEY_SOLVER)
        api.COLUMN_MAP_SOLVER = config.get('column_map_solver', api.COLUMN_MAP_SOLVER)
        api.COLUMN_MAP_THRESHOLD = config.get('column_map_threshold', api.COLUMN_MAP_THRESHOLD)

        host = config.get('host', args.host)
        port = config.get('port', args.port)
    else:
        host = args.host
        port = args.port

    print_box('Starting the DataTracer REST API', True)
    hug.API(api).http.serve(host, port, display_intro=False)


def _get_parser():
    # Logging
    logging_args = argparse.ArgumentParser(add_help=False)
    logging_args.add_argument('-v', '--verbose', action='count', default=0)

    parser = argparse.ArgumentParser(
        prog='datatracer',
        description='DataTracer Command Line Interface',
        parents=[logging_args]
    )

    subparsers = parser.add_subparsers(title='command', help='Command to execute')
    parser.set_defaults(command=None)

    api_parser = subparsers.add_parser(
        'api',
        parents=[logging_args],
        help='Start a development server for the DataTracer REST API.'
    )
    api_parser.set_defaults(command=_api_serve)
    api_parser.add_argument(
        '--host', default='127.0.0.1',
        help='IP at which to listen to. Defaults to 127.0.0.1. Use 0.0.0.0 for all interfaces.'
    )
    api_parser.add_argument(
        '-p', '--port', default=8000, type=int,
        help='Port at which to listen to. Defaults to 8000.'
    )
    api_parser.add_argument(
        '-c', '--config',
        help='Path to a configuration JSON or YAML file.'
    )

    return parser


def main():
    # Parse args
    parser = _get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    # Logger setup
    log_level = (3 - args.verbose) * 10
    fmt = '%(asctime)s - %(process)d - %(levelname)s - %(name)s - %(module)s - %(message)s'
    logging.basicConfig(level=log_level, format=fmt)

    args.command(args)


if __name__ == '__main__':
    main()



"""Foreign Key Solving base class."""


class ForeignKeySolver():

    def fit(self, list_of_databases):
        """Fit this solver.

        Args:
            list_of_databases (list):
                List of tuples containing ``MetaData`` instnces and table dictinaries,
                which contain table names as input and ``pandas.DataFrames`` as values.
        """
        pass

    def solve(self, tables, primary_keys=None):
        """Solve the foreign key detection problem.

        The output is a list of foreign key specifications, in order from the most likely
        to the least likely.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.
            primary_keys (dict):
                (Optional). Dictionary of table primary keys, as returned by the Primary
                Key Solvers. This parameter is optional and not all the subclasses need it.

        Returns:
            dict:
                List of foreign key specifications, sorted by likelyhood.
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""DataTracer core module.

This module defines the DataTracer class.
"""

import json
import os
import pickle

from mlblocks import MLPipeline

PRETRAINED_DIR = os.path.join(os.path.dirname(__file__), 'pretrained')


class DataTracer:
    """DataTracer Class.

    The DataTracer Class provides a unified and standardized access to
    all the Data Lineage functionalities of the project.

    Args:
        pipeline (str, dict or MLPipeline):
            Pipeline to use. It can be passed as:
                * An ``str`` with a path to a JSON file.
                * An ``str`` with the name of a registered pipeline.
                * An ``MLPipeline`` instance.
                * A ``dict`` with an ``MLPipeline`` specification.
        hyperparameters (dict):
            Additional hyperparameters to set to the Pipeline.
    """

    def _get_mlpipeline(self):
        pipeline = self._pipeline
        if isinstance(pipeline, str) and os.path.isfile(pipeline):
            with open(pipeline) as json_file:
                pipeline = json.load(json_file)

        mlpipeline = MLPipeline(pipeline)
        if self._hyperparameters:
            mlpipeline.set_hyperparameters(self._hyperparameters)

        return mlpipeline

    def __init__(self, pipeline, hyperparameters=None):
        self._pipeline = pipeline
        self._hyperparameters = hyperparameters
        self._mlpipeline = self._get_mlpipeline()

    def fit(self, datasets):
        """Fit the pipeline to the given data.

        Args:
            datasets (list or dict):
                List or dict of tuples containing a MetaData instance and a dict
                with the tables of the dataset loaded as DataFrames.
        """
        if isinstance(datasets, dict):
            datasets = list(datasets.values())

        self._mlpipeline = self._get_mlpipeline()
        self._mlpipeline.fit(list_of_databases=datasets, tables={})

    def solve(self, tables, **kwargs):
        """Solve the data lineage problem.

        The underlaying pipeline is executed and the outputs are returned.

        Args:
            tables (dict):
                Dictionary of tables from a dataset loaded as DataFrames.
            **kwargs:
                Any additional keyword arguments are passed down to the
                pipeline.

        Returns:
            object:
                This method returns the outputs returned by the pipeline.
        """
        return self._mlpipeline.predict(tables=tables, **kwargs)

    def save(self, path: str):
        """Save this object using pickle for later usage.

        Args:
            path (str):
                Path to the file where the serialization of
                this object will be stored.
        """
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as pickle_file:
            pickle.dump(self, pickle_file)

    @classmethod
    def load(cls, datatracer: str):
        """Load a DataTracer instance from a pickle file.

        Args:
            datatracer (str):
                Name of the datatracer to load, or path to a previously saved
                instance.

        Returns:
            DataTracer

        Raises:
            ValueError:
                If the serialized object is not a DataTracer instance.
        """
        if os.path.isfile(datatracer):
            path = datatracer
        else:
            path = os.path.join(PRETRAINED_DIR, datatracer + '.dt')
            if not os.path.isfile(path):
                raise ValueError('Unknown datatracer: {}'.format(datatracer))

        with open(path, 'rb') as pickle_file:
            datatracer = pickle.load(pickle_file)
            if not isinstance(datatracer, cls):
                raise ValueError('Serialized object is not a DataTracer instance')

            return datatracer



"""Column Mapping base class."""


class ColumnMapSolver:
    """Base Solver for the data lineage problem of column dependency."""

    def fit(self, list_of_databases):
        """Fit this solver.

        Args:
            list_of_databases (list):
                List of tuples containing ``MetaData`` instnces and table dictinaries,
                which contain table names as input and ``pandas.DataFrames`` as values.
        """
        pass

    def solve(self, tables, foreign_keys, target_table, target_field):
        """Find the fields which contributed to the target_field the most.

        The output is a dictionary containing the fields that contributed the
        most to the given target field as keys, specified as a tuple containing
        both table name and field name, and the score obtained as values.

        Args:
            tables (dict):
                Dict containing table names as input and ``pandas.DataFrames``
                as values.
            foreign_keys (list):
                List of foreign key specifications.
            target_table (str):
                Name of the table that contains the target field.
            target_field (str):
                Name of the target field.

        Returns:
            dict:
                Dictionary of field specification tuples and scores.
        """
        raise NotImplementedError()



from datatracer.foreign_key.base import ForeignKeySolver
from datatracer.foreign_key.basic import BasicForeignKeySolver
from datatracer.foreign_key.standard import StandardForeignKeySolver

__all__ = [
    'ForeignKeySolver',
    'BasicForeignKeySolver',
    'StandardForeignKeySolver'
]



import numpy as np


class Transformer:

    def __init__(self, tables, foreign_keys):
        """
        The `Transformer` class provides an interface between the database and
        the column mapping solver.

        For example, during the forwards pass, it could take a date field and
        transform into three columns - day, month, and year - which will allow
        the solver to potentially identify the lineage of some target column.

        Then, during the backwards pass, it can take the scores produced by the
        solver (i.e. the solver assigns scores indicating how important the day,
        month, and year columns are to predicting the target column) and combine
        them into a single score for the `date` field.
        """
        self.tables = tables
        self.foreign_keys = foreign_keys

    def forward(self, table, field):
        """
        This function returns a (X, y) tuple containing numerical values which
        is suitable for machine learning libraries. The `X` array contains
        data from columns that are potentially related to the target field. The
        `y` array contains the values of the target field.
        """
        df = self.tables[table]
        df = df.select_dtypes("number")
        df = df.fillna(0.0)
        X, y = df.drop([field], axis=1), df[field]
        self.columns = [(table, col_name) for col_name in X.columns]
        X, y = X.values, y.values

        X_new, columns_new = self._get_counts(table)
        if columns_new:
            X = np.concatenate([X, X_new], axis=1)
            self.columns.extend(columns_new)

        X_new, columns_new = self._get_aggregations(table)
        if columns_new:
            X = np.concatenate([X, X_new], axis=1)
            self.columns.extend(columns_new)

        return X, y

    def _get_counts(self, table):
        """
        Get the foreign keys where the given table is the parent.
        """
        X, columns = [], []
        for fk in self.foreign_keys:
            if fk["ref_table"] != table:
                continue

            # Count the number of rows for each key.
            child_table = self.tables[fk["table"]].copy()
            child_table["_dummy_"] = 0.0
            child_counts = child_table.groupby(fk["field"]).count().iloc[:, 0:1]
            child_counts.columns = ["_tmp_"]

            # Merge the counts into the parent table
            parent_table = self.tables[table]
            parent_table = parent_table.set_index(fk["ref_field"])
            parent_table = parent_table.join(child_counts).reset_index()

            X.append(parent_table["_tmp_"].fillna(0.0).values)
            columns.append((fk["table"], fk["field"]))

        return np.array(X).transpose(), columns

    def _get_aggregations(self, table):
        """
        Get the foreign keys where the given table is the parent.
        """
        X, columns = [], []
        for fk in self.foreign_keys:
            if fk["ref_table"] != table:
                continue

            for op, op_name in [
                (lambda x: x.sum(), "SUM"),
                (lambda x: x.max(), "MAX"),
                (lambda x: x.min(), "MIN"),
                (lambda x: x.std(), "STD"),
            ]:
                # Count the number of rows for each key.
                child_table = self.tables[fk["table"]].copy()
                if len(child_table.columns) <= 1:
                    continue

                child_counts = op(child_table.groupby(fk["field"]))
                old_column_names = list(child_counts.columns)
                child_counts.columns = ["%s(%s)" % (op_name, col_name)
                                        for col_name in old_column_names]

                # Merge the counts into the parent table
                parent_table = self.tables[table]
                parent_table = parent_table.set_index(fk["ref_field"])
                parent_table = parent_table.join(child_counts).reset_index()

                for old_name, col_name in zip(old_column_names, child_counts.columns):
                    if parent_table[col_name].dtype.kind == "f":
                        X.append(parent_table[col_name].fillna(0.0).values)
                        columns.append((fk["table"], old_name))

        return np.array(X).transpose(), columns

    def backward(self, feature_importances):
        """
        This function takes an array of `feature_importances` which corresponds
        to the `X` matrix produced by the last call to `forward`. It returns a
        mapping from fields to importance scores.
        """
        obj = {}
        for column, importance in zip(self.columns, feature_importances):
            obj[column] = importance

        return obj
